#!/usr/bin/env python

from scm.plams.core.settings import Settings
import numpy
from scm.plams.mol.atom import Atom
from scm.plams.mol.molecule import Molecule
from scm.plams.tools.periodic_table import PT
from scm.plams.trajectories.xyzfile import XYZTrajectoryFile, data_from_xyzcomment

__all__ = ["XYZHistoryFile"]


class XYZHistoryFile(XYZTrajectoryFile):
    """
    Class representing an XYZ file containing a molecular simulation history with varying numbers of atoms

    An instance of this class has the following attributes:

    *   ``file_object`` -- A Python :py:class:`file` object, referring to the actual XYZ file
    *   ``position``    -- The frame to which the cursor is currently pointing in the XYZ file
    *   ``mode``        -- Designates whether the file is in read or write mode ('r' or 'w')
    *   ``elements``    -- The elements of the atoms in the system at the current frame

    An |XYZHistoryFile| object behaves very similar to a regular file object.
    It has read and write methods (:meth:`read_next` and :meth:`write_next`)
    that read and write from/to the position of the cursor in the ``file_object`` attribute.
    If the file is in read mode, an additional method :meth:`read_frame` can be used that moves
    the cursor to any frame in the file and reads from there.
    The amount of information stored in memory is kept to a minimum, as only information from the current frame
    is ever stored.

    Reading and writing to and from the files can be done as follows::

        >>> from scm.plams import XYZHistoryFile

        >>> xyz = XYZHistoryFile('old.xyz')
        >>> mol = xyz.get_plamsmol()

        >>> xyzout = XYZHistoryFile('new.xyz',mode='w')

        >>> for i in range(xyz.get_length()) :
        >>>     crd,cell = xyz.read_frame(i,molecule=mol)
        >>>     xyzout.write_next(molecule=mol)

    The above script reads information from the XYZ file ``old.xyz`` into the |Molecule| object ``mol``
    in a step-by-step manner.
    The |Molecule| object is then passed to the :meth:`write_next` method of the new |XYZHistoryFile|
    object corresponding to the new xyz file ``new.xyz``.

    The exact same result can also be achieved by iterating over the instance as a callable

        >>> xyz = XYZHistoryFile('old.xyz')
        >>> mol = xyz.get_plamsmol()

        >>> xyzout = XYZHistoryFile('new.xyz',mode='w')

        >>> for crd,cell in xyz(mol) :
        >>>     xyzout.write_next(molecule=mol)

    This procedure requires all coordinate information to be passed to and from the |Molecule| object
    for each frame, which can be time-consuming.
    It is therefore also possible to bypass the |Molecule| object when reading through the frames::

        >>> xyz = XYZHistoryFile('old.xyz')

        >>> xyzout = XYZHistoryFile('new.xyz',mode='w')

        >>> for crd,cell in xyz :
        >>>     xyzout.write_next(coords=crd,elements=xyz.elements)

    By default the write mode will create a minimal version of the XYZ file, containing only elements
    and coordinates.
    Additional information can be written to the file by supplying additional arguments
    to the :meth:`write_next` method.
    The additional keywords `step` and `energy` trigger the writing of a remark containing
    the molecule name, the step number, the energy, and the lattice vectors.

        >>> mol = Molecule('singleframe.xyz')

        >>> xyzout = XYZHistoryFile('new.xyz',mode='w')
        >>> xyzout.set_name('MyMol')

        >>> xyzout.write_next(molecule=mol, step=0, energy=5.)
    """

    def __init__(self, filename, mode="r", fileobject=None, ntap=None):
        """
        Initiates an XYZHistoryFile object

        * ``filename``   -- The path to the XYZ file
        * ``mode``       -- The mode in which to open the XYZ file ('r' or 'w')
        * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
        * ``ntap``       -- If the file is in write mode, the number of atoms can be passed here
        """
        XYZTrajectoryFile.__init__(self, filename, mode, fileobject, ntap)

        self.input_elements = self.elements[:]

    def _is_endoffile(self):
        """
        If the end of file is reached, return coords and cell as None
        """
        end = False
        line = self.file_object.readline()
        if len(line) == 0:
            end = True
            return end
        nats = int(line.split()[0])
        for i in range(nats + 1):
            line = self.file_object.readline()
            if len(line) == 0:
                end = True
                break
        for i in range(self.nveclines):
            line = self.file_object.readline()
            if len(line) == 0:
                end = True
                break
        return end

    def _read_coordinates(self, molecule):
        """
        Read the coordinates at current step
        """
        # Find the number of atoms
        line = self.file_object.readline()
        if len(line) == 0:
            return None, None  # End of file is reached
        nats = int(line.split()[0])
        line = self.file_object.readline()

        # Handle the comment line
        cell = None
        historydata = data_from_xyzcomment(line)
        if "Lattice" in historydata:
            cell = historydata["Lattice"]
            del historydata["Lattice"]
        if self.include_historydata:
            self.historydata = historydata

        # Read coordinates and elements
        coords = []
        elements = []
        for i in range(nats):
            line = self.file_object.readline()
            words = line.split()
            coords.append([float(w) for w in words[1:4]])
            elements.append(words[0])

        # If the elements changed, update the molecule
        if elements != self.elements:
            self.elements = elements
            self.coords = numpy.array(coords)
            # Rebuild the molecule (bonds will disappear for now)
            if isinstance(molecule, Molecule):
                for at in reversed(molecule.atoms):
                    molecule.delete_atom(at)
                molecule.properties = Settings()
                for el in elements:
                    atom = Atom(PT.get_atomic_number(el))
                    molecule.add_atom(atom)
        else:
            self.coords[:] = coords

        # Possibly read lattice
        lattice = []
        for i in range(self.nveclines):
            line = self.file_object.readline()
            words = line.split()
            lattice.append([float(w) for w in words[1:]])
        if cell is None and len(lattice) > 0:
            cell = lattice

        # Assign the data to the molecule object
        if isinstance(molecule, Molecule):
            self._set_plamsmol(self.coords, cell, molecule)

        return coords, cell

    def write_next(
        self, coords=None, molecule=None, elements=None, cell=[0.0, 0.0, 0.0], conect=None, historydata=None
    ):
        """
        Write frame to next position in trajectory file

        * ``coords``   -- A list or numpy array of (``ntap``,3) containing the system coordinates
        * ``molecule`` -- A molecule object to read the molecular data from
        * ``elements`` -- The element symbols of the atoms in the system
        * ``cell``     -- A set of lattice vectors or cell diameters
        * ``energy``   -- An energy value to be written to the remark line
        * ``conect``   -- A dictionary containing connectivity info (not used)
        * ``historydata`` -- A dictionary containing additional variables to be written to the comment line

        The ``historydata`` dictionary can contain for example:
        ('Step','Energy'), the frame number and the energy respectively

        .. note::

                Either ``coords`` and ``elements`` or ``molecule`` are mandatory arguments
        """
        if isinstance(molecule, Molecule):
            coords, cell, elements = self._read_plamsmol(molecule)[:3]
        self.elements = elements
        cell = self._convert_cell(cell)

        self._write_moldata(coords, cell, historydata)

        self.position += 1



import collections
import math

from scm.plams.core.errors import UnitsError
import numpy as np

__all__ = ["Units"]


class Units:
    """A singleton class for unit converter.

    All values are based on `2014 CODATA recommended values <http://physics.nist.gov/cuu/Constants>`_.

    The following constants and units are supported:

    *   constants:

        -   ``speed_of_light`` (also ``c``)
        -   ``electron_charge`` (also ``e``)
        -   ``Avogadro_constant`` (also ``NA``)
        -   ``Bohr_radius``

    *   distance:

        -   ``Angstrom``, ``A``, ``Ang``
        -   ``Bohr``, ``au``, ``a.u.``
        -   ``nm``
        -   ``pm``

    *   reciprocal distance:

        -   ``1/Angstrom``, ``1/A``, ``Angstrom^-1``, ``A^-1``,
        -   ``1/Bohr``, ``Bohr^-1``

    *   angle:

        -    ``degree``, ``deg``
        -    ``radian``, ``rad``
        -    ``grad``
        -    ``circle``

    *   charge:

        -    ``coulomb``, ``C``
        -    ``e``

    *   energy:

        -   ``au``, ``a.u.``, ``Hartree``
        -   ``eV``
        -   ``kcal/mol``
        -   ``kJ/mol``
        -   ``cm^-1``, ``cm-1``
        -   ``K``, ``Kelvin``

    *   dipole moment:

        -   ``au``, ``a.u.``, ``e*bohr``
        -   ``Debye``, ``D``
        -  All charge units multiplied by distance units, for example
        -   ``eA``,  ``e*A``
        -   ``Cm``,  ``C*m``

    *   molecular polarizability:

        -   ``au``, ``a.u.``, ``(e*bohr)^2/hartree``
        -   ``e*A^2/V``
        -   ``C*m^2/V``
        -   ``cm^3``
        -   ``bohr^3``
        -   ``A^3``, ``angstrom^3``, ``Ang^3``

    *   forces:

        -   All energy units divided by angstrom or bohr, for example
        -   ``eV/angstrom``
        -   ``hartree/bohr``

    *   hessian:

        -   All energy units divided by angstrom^2 or bohr^2, for example
        -   ``eV/angstrom^2``
        -   ``hartree/bohr^2``

    *   pressure:

        -   All energy units divided by angstrom^3 or bohr^3, for example
        -   ``eV/angstrom^3``
        -   ``hartree/bohr^3``
        -   And some more:
        -   ``Pa``
        -   ``GPa``
        -   ``bar``
        -   ``atm``


    Example::

        >>> print(Units.constants['speed_of_light'])
        299792458
        >>> print(Units.constants['e'])
        1.6021766208e-19
        >>> print(Units.convert(123, 'angstrom', 'bohr'))
        232.436313431
        >>> print(Units.convert([23.32, 145.0, -34.7], 'kJ/mol', 'kcal/mol'))
        [5.573613766730401, 34.655831739961755, -8.293499043977056]
        >>> print(Units.conversion_ratio('kcal/mol', 'kJ/mol'))
        4.184


    """

    constants = {}
    constants["Bohr_radius"] = 0.529177210903  # A     http://physics.nist.gov/cgi-bin/cuu/Value?bohrrada0
    constants["Avogadro_constant"] = constants["NA"] = (
        6.022140857e23  # 1/mol http://physics.nist.gov/cgi-bin/cuu/Value?na
    )
    constants["speed_of_light"] = constants["c"] = 299792458  # m/s http://physics.nist.gov/cgi-bin/cuu/Value?c
    constants["electron_charge"] = constants["e"] = 1.6021766208e-19  # C http://physics.nist.gov/cgi-bin/cuu/Value?e
    constants["Boltzmann"] = constants["k_B"] = 1.380649e-23  # J/K
    constants["vacuum_electric_permittivity"] = (
        8.8541878128e-12  # F*m-1=C/(V*m) https://physics.nist.gov/cgi-bin/cuu/Value?ep0
    )

    distance = {}
    distance["A"] = distance["Angstrom"] = distance["Ang"] = 1.0
    distance["Bohr"] = distance["bohr"] = distance["a.u."] = distance["au"] = 1.0 / constants["Bohr_radius"]
    distance["nm"] = distance["A"] / 10.0
    distance["pm"] = distance["A"] * 100.0
    distance["m"] = distance["A"] * 1e-10

    rec_distance = {}
    rec_distance["1/A"] = rec_distance["1/Ang"] = rec_distance["1/Angstrom"] = rec_distance["A^-1"] = rec_distance[
        "Ang^-1"
    ] = rec_distance["Angstrom^-1"] = 1.0
    rec_distance["1/m"] = rec_distance["m^-1"] = 1e10
    rec_distance["1/Bohr"] = rec_distance["Bohr^-1"] = constants["Bohr_radius"]

    energy = {}
    energy["au"] = energy["a.u."] = energy["Hartree"] = energy["Ha"] = 1.0
    energy["eV"] = 27.211386245988  # http://physics.nist.gov/cgi-bin/cuu/Value?hrev
    energy["kJ/mol"] = 4.359744650e-21 * constants["NA"]  # http://physics.nist.gov/cgi-bin/cuu/Value?hrj
    energy["J"] = 4.359744650e-18
    energy["kcal/mol"] = energy["kJ/mol"] / 4.184
    energy["cm^-1"] = energy["cm-1"] = 219474.6313702  # http://physics.nist.gov/cgi-bin/cuu/Value?hrminv
    energy["K"] = energy["J"] / constants["k_B"]

    mass = {}
    mass["au"] = mass["a.u."] = mass["amu"] = 1.0
    mass["kg"] = 1.66053906660e-27
    mass["g"] = mass["kg"] * 1e3

    time = {}
    time["s"] = 1.0
    time["ms"] = time["s"] * 1e3
    time["us"] = time["s"] * 1e6
    time["ns"] = time["s"] * 1e9
    time["ps"] = time["s"] * 1e12
    time["fs"] = time["s"] * 1e15
    time["au"] = time["a.u."] = time["s"] / 2.4188843265857e-17  # https://physics.nist.gov/cgi-bin/cuu/Value?aut

    angle = {}
    angle["degree"] = angle["deg"] = 1.0
    angle["radian"] = angle["rad"] = math.pi / 180.0
    angle["grad"] = 100.0 / 90.0
    angle["circle"] = 1.0 / 360.0

    charge = {}
    charge["a.u."] = charge["au"] = charge["e"] = 1
    charge["C"] = charge["coulomb"] = constants["e"]

    dipole = {}
    for k, v in charge.items():
        if (k == "au") or (k == "a.u."):  # remove 'au','a.u.' options
            continue
        for k1, v1 in distance.items():
            if (k1 == "au") or (k1 == "a.u."):
                continue
            dipole[k + "*" + k1] = v * v1
            dipole[k + k1] = v * v1
    dipole["au"] = dipole["a.u."] = dipole["e*bohr"]
    dipole["debye"] = dipole["D"] = dipole["Cm"] * constants["c"] * 1e21

    # from support info https://doi.org/10.48550/arXiv.2310.13310 it is preferable to highlight that this is molecular polarizability,
    # it should be also /mol units but it is usually omitted and for consistency in the dipole units I removed, but both dipole and molecular_polarizability should have /mol
    molecular_polarizability = {}
    molecular_polarizability["au"] = molecular_polarizability["a.u."] = molecular_polarizability[
        "e^2*bohr^2/hartree"
    ] = molecular_polarizability["(e*bohr)^2/hartree"] = 1.0
    molecular_polarizability["e*A^2/V"] = molecular_polarizability["e^2*A^2/eV"] = molecular_polarizability[
        "(e*A)^2/eV"
    ] = (constants["Bohr_radius"] ** 2 / energy["eV"])
    molecular_polarizability["C*m^2/V"] = molecular_polarizability["e*A^2/V"] * 1e-20 * constants["e"]
    molecular_polarizability["cm^3"] = (
        molecular_polarizability["C*m^2/V"] / (4 * np.pi * constants["vacuum_electric_permittivity"]) * 1e6
    )  # form https://en.wikipedia.org/wiki/Polarizability that refs Atkins book
    molecular_polarizability["A^3"] = molecular_polarizability["Ang^3"] = molecular_polarizability["Angstrom^3"] = (
        molecular_polarizability["cm^3"] * 1e24
    )
    molecular_polarizability["bohr^3"] = molecular_polarizability["Ang^3"] / constants["Bohr_radius"] ** 3

    forces = {}
    hessian = {}
    stress = {}
    for k, v in energy.items():
        for k1, v1 in distance.items():
            forces[k + "/" + k1] = v / v1
            hessian[k + "/" + k1 + "^2"] = v / v1**2
            stress[k + "/" + k1 + "^3"] = v / v1**3
    forces["au"] = forces["a.u."] = forces["Ha/bohr"]
    hessian["au"] = hessian["a.u."] = hessian["Ha/bohr^2"]
    stress["au"] = stress["a.u."] = stress["Ha/bohr^3"]
    stress["Pa"] = stress["J/m^3"]
    stress["GPa"] = stress["Pa"] * 1e-9
    stress["bar"] = stress["Pa"] * 1e-5
    stress["atm"] = stress["bar"] / 1.01325

    dicts = {}
    dicts["distance"] = distance
    dicts["energy"] = energy
    dicts["mass"] = mass
    dicts["time"] = time
    dicts["angle"] = angle
    dicts["dipole"] = dipole
    dicts["reciprocal distance"] = rec_distance
    dicts["forces"] = forces
    dicts["hessian"] = hessian
    dicts["stress"] = stress
    dicts["charge"] = charge
    dicts["molecular_polarizability"] = molecular_polarizability

    # Precomputed a dict mapping lowercased unit names to quantityName:conversionFactor pairs
    quantities_for_unit = {}
    for quantity in dicts:
        for unit, factor in dicts[quantity].items():
            unit = unit.lower()
            if unit not in quantities_for_unit:
                quantities_for_unit[unit] = {}
            quantities_for_unit[unit][quantity] = factor

    def __init__(self):
        raise UnitsError("Instances of Units cannot be created")

    @classmethod
    def find_unit(cls, unit):
        ret = {}
        u = unit.lower()
        quantities = cls.quantities_for_unit.get(u, {})
        for quantity in quantities:
            for k in cls.dicts[quantity]:
                if k.lower() == u:
                    ret[quantity] = k
                    break
        return ret

    @classmethod
    def conversion_ratio(cls, inp, out) -> float:
        """Return conversion ratio from unit *inp* to *out*."""
        if inp == out:
            return 1.0
        inps = cls.quantities_for_unit.get(inp.lower(), {})
        outs = cls.quantities_for_unit.get(out.lower(), {})
        common = set(inps.keys()) & set(outs.keys())
        if len(common) > 0:
            quantity = common.pop()
            return outs[quantity] / inps[quantity]
        else:
            if len(inps) == 0 and len(outs) == 0:
                raise UnitsError("Unsupported units: '{}' and '{}'".format(inp, out))
            if len(inps) > 0 and len(outs) > 0:
                raise UnitsError(
                    "Invalid unit conversion: '{}' is a unit of {} and '{}' is a unit of {}".format(
                        inp, ", ".join(list(inps.keys())), out, ", ".join(list(outs.keys()))
                    )
                )
            else:  # exactly one of (inps,outs) empty
                invalid, nonempty = (out, inps) if len(inps) else (inp, outs)
                if len(nonempty) == 1:
                    quantity = list(nonempty.keys())[0]
                    raise UnitsError(
                        "Invalid unit conversion: {} is not supported. Supported units for {}: {}".format(
                            invalid, quantity, ", ".join(list(cls.dicts[quantity].keys()))
                        )
                    )
                else:
                    raise UnitsError(
                        "Invalid unit conversion: {} is not a supported unit for {}".format(
                            invalid, ", ".join(list(nonempty.keys()))
                        )
                    )

    @classmethod
    def convert(cls, value, inp, out):
        """Convert *value* from unit *inp* to *out*.

        *value* can be a single number or a container (list, tuple, numpy.array etc.). In the latter case a container of the same type and length is returned. Conversion happens recursively, so this method can be used to convert, for example, a list of lists of numbers, or any other hierarchical container structure. Conversion is applied on all levels, to all values that are numbers (also numpy number types). All other values (strings, bools etc.) remain unchanged.
        """
        if value is None or isinstance(value, (bool, str)) or inp == out:
            return value
        if isinstance(value, collections.abc.Iterable):
            t = type(value)
            if t == np.ndarray:
                t = np.array
            v = [cls.convert(i, inp, out) for i in value]
            return t(v)
        if isinstance(value, (int, float, np.generic)):
            return value * cls.conversion_ratio(inp, out)
        return value

    @classmethod
    def ascii2unicode(cls, string):
        """
        Converts '^2' to '²' etc., for prettier printing of units.
        """
        if string is None:
            return ""
        ret = (
            string.replace("^-1", "⁻¹")
            .replace("angstrom", "Å")
            .replace("^2", "²")
            .replace("^3", "³")
            .replace("degree", "°")
            .replace("deg.", "°")
            .replace("Ang", "Å")
            .replace("*", "⋅")
        )
        return ret



#!/usr/bin/env amspython
from ase.optimize import BFGS
from scm.plams import *
from scm.plams.interfaces.adfsuite.ase_calculator import AMSCalculator


def get_atoms():
    # water in a box
    mol = from_smiles("O")  # PLAMS Molecule
    mol.lattice = [
        [
            3.0,
            0.0,
            0.0,
        ],
        [0.0, 3.0, 0.0],
        [0.0, 0.0, 3.0],
    ]
    return toASE(mol)  # convert PLAMS Molecule to ASE Atoms


def get_settings():
    # PLAMS Settings configuring the calculation
    s = Settings()
    s.input.ams.Task = "SinglePoint"
    s.input.ams.Properties.Gradients = "Yes"
    s.input.ams.Properties.StressTensor = "Yes"

    # Engine definition
    s.input.ForceField.Type = "UFF"

    # run in serial
    s.runscript.nproc = 1
    return s


def singlepoint():
    settings = get_settings()
    atoms = get_atoms()
    atoms.calc = AMSCalculator(settings=settings, name="SinglePoint")
    print("Singlepoint through the ASE calculator")
    print(f"Energy (eV): {atoms.get_potential_energy()}")
    print("Forces (eV/ang):")
    print(atoms.get_forces())
    print("Stress (eV/ang^3):")
    print(atoms.get_stress())


def ams_geoopt():
    print("AMS geo opt run with the ASE calculator")
    settings = Settings()
    settings.input.ams.Task = "GeometryOptimization"
    settings.input.ams.GeometryOptimization.Convergence.Gradients = 0.01  # hartree/ang
    settings.input.ForceField.Type = "UFF"
    settings.runscript.nproc = 1
    atoms = get_atoms()
    atoms.calc = AMSCalculator(settings=settings, name="AMS_GeoOpt")
    print(f"Optimized energy (eV): {atoms.get_potential_energy()}")


def ase_geoopt():
    print("ASE geo opt (ase.optimize.BGFGS) in normal mode: One results dir is saved for every step")
    settings = get_settings()
    atoms = get_atoms()
    atoms.calc = AMSCalculator(settings=settings, name="ASE_GeoOpt")
    dyn = BFGS(atoms)
    dyn.run(fmax=0.27)
    print(f"Optimized energy (eV): {atoms.get_potential_energy()}")


def ase_geoopt_workermode():
    print("ASE geo opt (ase.optimize.FGS) in AMSWorker mode: no output files are saved, minimal overhead")
    settings = get_settings()
    atoms = get_atoms()
    with AMSCalculator(settings=settings, name="ASE_WorkerGeoOpt", amsworker=True) as calc:
        atoms.calc = calc
        dyn = BFGS(atoms)
        dyn.run(fmax=0.27)
        print(f"Optimized energy (eV): {atoms.get_potential_energy()}")


def charged_system():
    print("Define the charge of a system through the ASE atoms object")
    settings = get_settings()
    atoms = get_atoms()
    atoms.set_initial_charges([-0.3, 0.2, 0.2])
    atoms.info["charge"] = 3
    calc = AMSCalculator(settings, name="charge")
    atoms.calc = calc
    atoms.get_potential_energy()
    print(f'Charge is "{atoms.calc.amsresults.get_input_molecule().properties.charge}" for the first system.')
    atoms = atoms * (1, 1, 2)
    atoms.calc = calc
    atoms.get_potential_energy()
    print(f'Charge is "{atoms.calc.amsresults.get_input_molecule().properties.charge}" for the second system.')


def main():
    init()
    singlepoint()
    ams_geoopt()
    ase_geoopt()
    ase_geoopt_workermode()
    charged_system()
    finish()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import io

from scm.plams.core.errors import TrajectoryError
import numpy
from scm.plams.mol.atom import Atom
from scm.plams.mol.bond import Bond
from scm.plams.mol.molecule import Molecule
from scm.plams.trajectories.trajectoryfile import TrajectoryFile

__all__ = ["SDFTrajectoryFile", "create_sdf_string"]


class SDFTrajectoryFile(TrajectoryFile):
    """
    Class representing an SDF file containing a molecular trajectory

    An instance of this class has the following attributes:

    *   ``file_object`` -- A Python :py:class:`file` object, referring to the actual SDF file
    *   ``position``    -- The frame to which the cursor is currently pointing in the SDF file
    *   ``mode``        -- Designates whether the file is in read or write mode ('r' or 'w')
    *   ``ntap``        -- The number of atoms in the molecular system (needs to be constant throughout)
    *   ``elements``    -- The elements of the atoms in the system (needs to be constant throughout)

    An |SDFTrajectoryFile| object behaves very similar to a regular file object.
    It has read and write methods (:meth:`read_next` and :meth:`write_next`)
    that read and write from/to the position of the cursor in the ``file_object`` attribute.
    If the file is in read mode, an additional method :meth:`read_frame` can be used that moves
    the cursor to any frame in the file and reads from there.
    The amount of information stored in memory is kept to a minimum, as only information from the current frame
    is ever stored.

    Reading and writing to and from the files can be done as follows::

        >>> from scm.plams import SDFTrajectoryFile

        >>> sdf = SDFTrajectoryFile('old.sdf')
        >>> mol = sdf.get_plamsmol()

        >>> sdfout = SDFTrajectoryFile('new.sdf',mode='w')

        >>> for i in range(sdf.get_length()) :
        >>>     crd,cell = sdf.read_frame(i,molecule=mol)
        >>>     sdfout.write_next(molecule=mol)

    The above script reads information from the SDF file ``old.sdf`` into the |Molecule| object ``mol``
    in a step-by-step manner.
    The |Molecule| object is then passed to the :meth:`write_next` method of the new |SDFTrajectoryFile|
    object corresponding to the new sdf file ``new.sdf``.

    The exact same result can also be achieved by iterating over the instance as a callable

        >>> sdf =SDFZTrajectoryFile('old.sdf')
        >>> mol = sdf.get_plamsmol()

        >>> sdfout = SDFTrajectoryFile('new.sdf',mode='w')

        >>> for crd,cell in sdf(mol) :
        >>>     sdfout.write_next(molecule=mol)

    This procedure requires all coordinate information to be passed to and from the |Molecule| object
    for each frame, which can be time-consuming.
    It is therefore also possible to bypass the |Molecule| object when reading through the frames::

        >>> sdf = SDFTrajectoryFile('old.sdf')

        >>> sdfout = SDFTrajectoryFile('new.sdf',mode='w')
        >>> sdfout.set_elements(sdf.get_elements())

        >>> for crd,cell in sdf :
        >>>     sdfout.write_next(coords=crd)
        >>> sdfout.close()

    By default the write mode will create a minimal version of the SDF file, containing only elements
    and coordinates.
    Additional information can be written to the file by supplying additional arguments
    to the :meth:`write_next` method.
    The additional keywords `step` and `energy` trigger the writing of a remark containing
    the molecule name, the step number, the energy, and the lattice vectors.

        >>> mol = Molecule('singleframe.xyz')

        >>> sdfout = SDFTrajectoryFile('new.sdf',mode='w')
        >>> sdfout.set_name('MyMol')

        >>> sdfout.write_next(molecule=mol, step=0, energy=5.)
    """

    def __init__(self, filename, mode="r", fileobject=None, ntap=None):
        """
        Initiates an SDFTrajectoryFile object

        * ``filename``   -- The path to the SDF file
        * ``mode``       -- The mode in which to open the SDF file ('r' or 'w')
        * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
        * ``ntap``       -- If the file is in write mode, the number of atoms needs to be passed here
        """
        TrajectoryFile.__init__(self, filename, mode, fileobject, ntap)

        # SDF specific attributes
        self.name = "PlamsMol"
        self.conect = None

        # Specific SDF stuff
        self.include_historydata = False
        self.historydata = None

        # Required setup before frames can be read/written
        if self.mode == "r":
            self._read_header()
        elif self.mode == "a":
            self._move_cursor_to_append_pos()

    def store_historydata(self):
        """
        Additional data should be read from/written to file
        """
        self.include_historydata = True

    def set_name(self, name):
        """
        Sets the name of the system, in case an extensive write is requested

        *   ``name`` -- A string containing the name of the molecule
        """
        self.name = name

    def _read_header(self):
        """
        Set up info required for reading frames
        """
        # Read the first molecule
        lines = []
        while True:
            line = self.file_object.readline()
            if len(line) == 0:
                raise Exception("End not found")
            lines.append(line)
            if len(line) < 6:
                continue
            if line[:6] == "M  END":
                break
        mol, _ = get_molecule(lines)
        self.ntap = len(mol)
        self.elements = [at.symbol for at in mol]
        if self.coords.shape == (0, 3):
            self.coords = numpy.zeros((self.ntap, 3))

        # Rewind
        self.file_object.seek(0)

    def read_next(self, molecule=None, read=True):
        """
        Reads coordinates from the current position of the cursor and returns it

        * ``molecule`` -- |Molecule| object in which the new coordinates need to be stored
        * ``read``     -- If set to False the cursor will move to the next frame without reading
        """
        if not read and not self.firsttime:
            return self._move_cursor_without_reading()

        cell = None
        # Read the coordinates
        crd, cell = self._read_coordinates(molecule)
        if crd is None:
            return None, None  # End of file is reached

        if self.firsttime:
            self.firsttime = False

        self.position += 1

        return self.coords, cell

    def _read_coordinates(self, molecule):
        """
        Read the coordinates from file, and place them in the molecule
        """
        # Read lines until the end
        lines = []
        while True:
            line = self.file_object.readline()
            if len(line) == 0:
                return None, None  # End of file is reached
            lines.append(line)
            if len(line) < 4:
                continue
            if line[:4] == "$$$$":
                break

        # Get the mol part
        mol, restlines = get_molecule(lines)
        if len(mol) != self.ntap:
            raise TrajectoryError("Number of atoms changes. Not implemented.")

        # Get the coordinates and cell
        cell = mol.lattice
        if len(cell) == 0:
            cell = None
        self.coords[:, :] = mol.as_array()
        if len(mol.bonds) > 0:
            conect = {}
            for bond in mol.bonds:
                iat = min(mol.index(bond))
                jat = max(mol.index(bond))
                if not iat in conect:
                    conect[iat] = []
                conect[iat].append((jat, bond.order))
            self.conect = conect

        # Read the additional data
        if self.include_historydata:
            historydata = {}
            # First find all entries (entries can run over multiple lines)
            entries = [i for i, line in enumerate(restlines[:-1]) if line[:4] == ">  <"] + [len(restlines) - 1]
            for i, iline in enumerate(entries[:-1]):
                key = restlines[iline].split("<")[1].split(">")[0]
                value = "".join(restlines[iline + 1 : entries[i + 1] - 1])
                value = value.strip()
                # Try to turn this into a float or integer?
                if value.isdigit():
                    value = int(value)
                else:
                    try:
                        value = float(value)
                    except ValueError:
                        pass
                historydata[key] = value
            self.historydata = historydata

        if isinstance(molecule, Molecule):
            self._set_plamsmol(self.coords, cell, molecule)

        return self.coords, cell

    def _is_endoffile(self):
        """
        If the end of file is reached, return coords and cell as None
        """
        end = False
        while True:
            line = self.file_object.readline()
            if len(line) == 0:
                end = True
                break
            if len(line) < 4:
                continue
            if line[:4] == "$$$$":
                break
        return end

    def write_next(self, coords=None, molecule=None, cell=[0.0, 0.0, 0.0], conect=None, historydata=None):
        """
        Write frame to next position in trajectory file

        * ``coords``   -- A list or numpy array of (``ntap``,3) containing the system coordinates
        * ``molecule`` -- A molecule object to read the molecular data from
        * ``cell``     -- A set of lattice vectors or cell diameters
        * ``conect``   -- A dictionary containing connectivity info (not used)
        * ``historydata`` -- A dictionary containing additional variables to be written to the comment line

        The ``historydata`` dictionary can contain for example:
        ('Step','Energy'), the frame number and the energy respectively

        .. note::

                Either ``coords`` or ``molecule`` are mandatory arguments
        """
        if isinstance(molecule, Molecule):
            coords, cell, elements = self._read_plamsmol(molecule)[:3]
            if self.position == 0:
                self.elements = elements
        cell = self._convert_cell(cell)

        if not isinstance(molecule, Molecule):
            # Create the molecule?
            molecule = Molecule()
            for el, crd in zip(self.elements, coords):
                atom = Atom(symbol=el, coords=crd)
                molecule.add_atom(atom)
            if cell is not None:
                molecule.lattice = cell.tolist()
            # Add the bonds
            bondlist = []
            if conect is not None:
                for iat, neighbors in conect.items():
                    for t in neighbors:
                        jat = t
                        if isinstance(t, tuple):
                            jat = t[0]
                        indices = tuple(sorted([iat, jat]))
                        if not indices in bondlist:
                            bondlist.append(indices)
                            bond = Bond(molecule[iat], molecule[jat])
                            molecule.add_bond(bond)

        self._write_moldata(molecule, historydata)

        self.position += 1

    def _write_moldata(self, molecule, historydata):
        """
        Write all molecular info to file
        """
        block = create_sdf_string(molecule, self.position, historydata)
        self.file_object.write(block)

    def _rewind_to_first_frame(self):
        """
        Rewind the file to the first frame
        """
        self.file_object.seek(0)
        self.firsttime = True
        self.position = 0

    def _rewind_n_frames(self, nframes):
        """
        Rewind the file by nframes frames
        """
        new_frame = self.position - nframes
        self._rewind_to_first_frame()
        for i in range(new_frame):
            self.read_next(read=False)


def create_sdf_string(molecule, step=None, historydata=None):
    """
    Write an SDF entry based on the elements and the coordinates of the atoms
    """
    energy = 0.0
    if historydata is not None:
        if "Energy" in historydata:
            energy = historydata["Energy"]
        # The conformer case
        elif "energies" in historydata:
            energy = historydata["energies"]
    else:
        historydata = {}

    if "Step" in historydata:
        step = historydata["Step"]

    block = "Energy = %.10f kcal/mol\n" % (energy)
    f = io.StringIO()
    molecule.writemol(f)
    f.seek(0)
    text = f.read()
    f.close()
    text = "\n".join(text.split("\n")[1:])
    block += text
    for key, item in historydata.items():
        text = ">  <%s>  (%i)\n" % (key, step)
        text += f"{item}\n\n"
        block += text
    block += "$$$$\n"
    return block


def get_molecule(lines):
    """
    Read a molecule object from the lines for a single frame
    """
    end = 0
    for i, line in enumerate(lines):
        if line[:6] == "M  END":
            end = i + 1
            break
    moltext = "".join(lines[:end])
    f = io.StringIO(moltext)
    mol = Molecule()
    mol.readmol(f)
    f.close()
    restlines = lines[end:]
    return mol, restlines



from scm.plams.core.basejob import MultiJob
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings
from scm.plams.interfaces.adfsuite.ams import AMSJob
from scm.plams.mol.molecule import Molecule

__all__ = ["ADFFragmentJob", "ADFFragmentResults"]


class ADFFragmentResults(Results):

    def get_properties(self):
        return self.job.full.results.get_properties()

    def get_main_molecule(self):
        return self.job.full.results.get_main_molecule()

    def get_input_molecule(self):
        return self.job.full.results.get_input_molecule()

    def get_energy(self, unit="au"):
        return self.job.full.results.get_energy(unit)

    def get_dipole_vector(self, unit="au"):
        return self.job.full.results.get_dipole_vector(unit)

    def get_energy_decomposition(self):
        energy_section = self.job.full.results.read_rkf_section("Energy", file="adf")
        ret = {}
        for k in ["Electrostatic Energy", "Kinetic Energy", "Elstat Interaction", "XC Energy"]:
            ret[k] = energy_section[k]
        return ret


class ADFFragmentJob(MultiJob):
    _result_type = ADFFragmentResults

    def __init__(self, fragment1=None, fragment2=None, full_settings=None, **kwargs):
        MultiJob.__init__(self, **kwargs)
        self.fragment1 = fragment1.copy() if isinstance(fragment1, Molecule) else fragment1
        self.fragment2 = fragment2.copy() if isinstance(fragment2, Molecule) else fragment2
        self.full_settings = full_settings or Settings()

    def prerun(self):  # noqa F811
        self.f1 = AMSJob(name="frag1", molecule=self.fragment1, settings=self.settings)
        self.f2 = AMSJob(name="frag2", molecule=self.fragment2, settings=self.settings)

        for at in self.fragment1:
            at.properties.suffix = "adf.f=subsystem1"
        for at in self.fragment2:
            at.properties.suffix = "adf.f=subsystem2"

        self.full = AMSJob(
            name="full", molecule=self.fragment1 + self.fragment2, settings=self.settings + self.full_settings
        )

        self.full.settings.input.adf.fragments.subsystem1 = (self.f1, "adf")
        self.full.settings.input.adf.fragments.subsystem2 = (self.f2, "adf")

        self.children = [self.f1, self.f2, self.full]



#!/usr/bin/env amspython
# coding: utf-8

# ## Initial imports

import scm.plams as plams
from scm.plams.tools.plot import plot_correlation, get_correlation_xy
import matplotlib.pyplot as plt


plams.init()


# ## Define two engines to compare
#
# Here we choose GFNFF and GFN1-xTB

e1 = plams.Settings()
e1.input.GFNFF

e2 = plams.Settings()
e2.input.DFTB.Model = "GFN1-xTB"


# Let's use a glycine molecule generated from SMILES:

glycine = plams.from_smiles("C(C(=O)O)N")
plams.plot_molecule(glycine)


# Run a single-point calculation storing the Gradients (negative forces):

sp = plams.Settings()
sp.input.ams.Task = "SinglePoint"
sp.input.ams.Properties.Gradients = "Yes"
sp.runscript.nproc = 1  # run in serial


job1 = plams.AMSJob(settings=sp + e1, name="glycine-engine1", molecule=glycine)
job2 = plams.AMSJob(settings=sp + e2, name="glycine-engine2", molecule=glycine)


job1.run()
job2.run()


plot_correlation(
    job1,
    job2,
    section="AMSResults",
    variable="Gradients",
    file="engine",
)


# To get the actual numbers, use ``get_correlation_xy``:

x, y = plams.tools.plot.get_correlation_xy(job1, job2, section="AMSResults", variable="Gradients", file="engine")
print("x")
print(x)
print("y")
print(y)


# ## Compare multiple jobs

smiles_list = ["CC=C", "CCCO", "C(C(=O)O)N"]
names = ["propene", "propanol", "glycine"]
molecules = [plams.from_smiles(x) for x in smiles_list]
for mol in molecules:
    plams.plot_molecule(mol)


jobs1 = [plams.AMSJob(settings=sp + e1, name="e1" + name, molecule=mol) for name, mol in zip(names, molecules)]
jobs2 = [plams.AMSJob(settings=sp + e2, name="e2" + name, molecule=mol) for name, mol in zip(names, molecules)]


for job in jobs1 + jobs2:
    job.run()


# The correlation plot can be plotted as before. You can also add a unit conversion to get your preferred units, and add custom xlabel and ylabel:

unit = "eV/angstrom"
multiplier = plams.Units.convert(1.0, "hartree/bohr", unit)

plot_correlation(
    jobs1,
    jobs2,
    section="AMSResults",
    variable="Gradients",
    file="engine",
    xlabel="Engine 1",
    ylabel="Engine 2",
    unit=unit,
    multiplier=multiplier,
)


plot_correlation(
    jobs1,
    jobs2,
    section="AMSResults",
    variable="Charges",
    file="engine",
    xlabel="Engine 1",
    ylabel="Engine 2",
)


# ## Use Task Replay to compare multiple frames from a trajectory

# The forces from an MD job can be stored with ``writeenginegradients=True``

md = plams.AMSNVEJob(
    settings=e1,
    name="nve-md-e1",
    molecule=glycine,
    velocities=400,
    nsteps=100,
    samplingfreq=10,
    writeenginegradients=True,
)
md.run()


#

# When using the Replay task, set ``Properties.Gradients`` to get the forces:

replay_s = plams.Settings()
replay_s.input.ams.Task = "Replay"
replay_s.input.ams.Properties.Gradients = "Yes"
replay_s.input.ams.Replay.File = md.results.rkfpath()
replay = plams.AMSJob(settings=e2 + replay_s, name="replay-e2")
replay.run()


# For the MD job the gradients (negative forces) are stored in ``History%EngineGradients``, but for the Replay job they are stored in ``History%Gradients``. Use the ``alt_variable`` to specify the variable for the second job:

plot_correlation(md, replay, section="History", variable="EngineGradients", alt_variable="Gradients", file="ams")



#!/usr/bin/env amspython
from scm.plams import *


def main():
    mol = packmol(from_smiles("CCO"), density=0.8, n_molecules=4)  # 4 ethanol molecules in cubic box

    # Start by running UFF with Berendsen thermostat
    s = Settings()
    s.runscript.nproc = 1
    s.input.forcefield.type = "UFF"
    pre_eq = AMSNVTJob(
        settings=s, molecule=mol, name="pre_eq", temperature=400, timestep=1.0, thermostat="berendsen", nsteps=5
    )
    assert pre_eq.settings.input.ams.MolecularDynamics.Thermostat.Type == "berendsen"  # explicitly set above
    assert (
        pre_eq.settings.input.ams.MolecularDynamics.Thermostat.Tau == 400
    )  # default 400 * timestep if not set explicitly
    pre_eq.run()

    # Then switch to GAFF with an NHC thermostat
    s.input.forcefield.type = "GAFF"
    s.input.forcefield.antechamberintegration = "Yes"
    eq = AMSNVTJob.restart_from(
        pre_eq, settings=s, name="eq", nsteps=100, thermostat="NHC", samplingfreq=10, writevelocities=True
    )
    assert eq.settings.input.ams.MolecularDynamics.Thermostat.Type == "NHC"  # explicitly set above
    assert eq.settings.input.ams.MolecularDynamics.TimeStep == 1.0  # inherited from pre_eq
    eq.run()

    # Launch NVE simulation from some frames (structures+velocities) of the previous NVT simulation (still using GAFF)
    for frame in [1, 4, 7]:
        nvejob = AMSNVEJob.restart_from(eq, name=f"nve{frame}", frame=frame, nsteps=10, samplingfreq=1)
        assert "Thermostat" not in nvejob.settings.input.ams.MolecularDynamics
        nvejob.run()

    # Also run an NPT simulation at a pressure of 2 bar
    npt = AMSNPTJob.restart_from(
        eq, name="npt", nsteps=50, samplingfreq=10, pressure=2, thermostat="NHC", barostat="MTK"
    )
    assert npt.settings.input.ams.MolecularDynamics.Thermostat.Temperature == 400  # inherited
    npt.run()

    # continue NVT
    nvt_from_npt = AMSNVTJob.restart_from(npt, name="nvt_from_npt")
    assert "Barostat" not in nvt_from_npt.settings.input.ams.MolecularDynamics
    assert nvt_from_npt.settings.input.ams.MolecularDynamics.NSteps == 50  # inherited
    nvt_from_npt.run()

    # NVE from NPT
    nve_from_npt = AMSNVEJob.restart_from(npt, name="nve_from_npt")
    assert "Barostat" not in nve_from_npt.settings.input.ams.MolecularDynamics
    assert "Thermostat" not in nve_from_npt.settings.input.ams.MolecularDynamics
    assert nve_from_npt.settings.input.ams.MolecularDynamics.NSteps == 50  # inherited
    nve_from_npt.run()

    # NVE with random initial velocities
    nve_clean = AMSNVEJob(name="nve_clean", settings=s, molecule=mol, velocities=200, nsteps=20)
    nve_clean.run()


if __name__ == "__main__":
    init()
    main()
    finish()



#!/usr/bin/env plams
from scm.conformers import ConformersJob, ConformersResults

# This example shows how to use the AMS's Conformers tool via PLAMS


def print_results(results: ConformersResults):
    for i, (energy, mol) in enumerate(zip(results.get_energies("kcal/mol"), results.get_conformers())):
        print(f"Energy conformer {i} = {energy} [kcal/mol]")
        print(mol)


ethanol = from_smiles("OCC")

# Simple example: generate conformers for ethanol using default settings
# - default Task: Generate
# - default Generator Method: RDKit
# - default Engine: ForceField (with UFF)

job = ConformersJob(name="conformers_generation", molecule=ethanol)
job.run()

print("Conformers generated using RDKit and UFF:")
print_results(job.results)

# Re-optimize the conformers generated in the previous steps using the GFN1-xTB engine:

sett = Settings()

# In the 'ams' part of the settings input you can specify
# the input options for the Conformers tool, which are
# described in the Conformers tool user manual.

sett.input.ams.Task = "Optimize"
sett.input.ams.InputConformersSet = job.results

# You can specify the engine to be used (and the engine options) like you would
# for an AMSJob. See the AMSJob documentation for more details.

sett.input.DFTB.Model = "GFN1-xTB"

# Note: here we do not specify the input molecule because we are passing the results
# of a previous ConformersJob (ConformersResults) via the "InputConformersSet" input.

reoptimize_job = ConformersJob(name="optimize_conformers", settings=sett)
reoptimize_job.run()

print("Conformers re-optimized using the more accurate GFN1-xTB method:")
print_results(reoptimize_job.results)

temperature = 273
print(f"Boltzmann distribution at {temperature} Kelvin")
for i, p in enumerate(reoptimize_job.results.get_boltzmann_distribution(temperature)):
    print(f"Conformer {i} probability: {p}")



#!/usr/bin/env plams
# Workflow with PLAMS

# A couple of useful function for extracting results:


@add_to_class(AMSResults)
def get_excitations(results):
    """Returns excitation energies (in eV) and oscillator strenghts (in Debye)."""
    if results.job.ok():
        exci_energies_au = results.readrkf("Excitations SS A", "excenergies", file="engine")
        oscillator_str_au = results.readrkf("Excitations SS A", "oscillator strengths", file="engine")
        # The results are stored in atomic units. Convert them to more convenient units:
        exci_energies = Units.convert(exci_energies_au, "au", "eV")
        oscillator_str = Units.convert(oscillator_str_au, "au", "Debye")
        return exci_energies, oscillator_str
    else:
        return [], []


@add_to_class(AMSResults)
def has_good_excitations(results, min_energy, max_energy, oscillator_str_threshold=1e-4):
    """Returns True if there is at least one excitation with non-vanishing oscillator strenght
    in the energy range [min_energy, max_energy]. Unit for min_energy and max energy: eV."""
    exci_energies, oscillator_str = results.get_excitations()
    for e, o in zip(exci_energies, oscillator_str):
        if min_energy < e < max_energy and o > oscillator_str_threshold:
            return True
    return False


# Calculation settings:
# =====================

# Settings for geometry optimization with the AMS driver:
go_sett = Settings()
go_sett.input.ams.Task = "GeometryOptimization"
go_sett.input.ams.GeometryOptimization.Convergence.Gradients = 1.0e-4

# Settings for single point calculation with the AMS driver
sp_sett = Settings()
sp_sett.input.ams.Task = "SinglePoint"

# Settings for the DFTB engine (including excitations)
dftb_sett = Settings()
dftb_sett.input.dftb.Model = "SCC-DFTB"
dftb_sett.input.dftb.ResourcesDir = "QUASINANO2015"
dftb_sett.input.dftb.Properties.Excitations.TDDFTB.calc = "singlet"
dftb_sett.input.dftb.Properties.Excitations.TDDFTB.lowest = 10
dftb_sett.input.dftb.Occupation.Temperature = 5.0

# Settings for the geometry optimization with the ADF engine
adf_sett = Settings()
adf_sett.input.adf.Basis.Type = "DZP"
adf_sett.input.adf.NumericalQuality = "Basic"

# Settings for the excitation calculation using the ADF engine
adf_exci_sett = Settings()
adf_exci_sett.input.adf.Basis.Type = "TZP"
adf_exci_sett.input.adf.XC.GGA = "PBE"
adf_exci_sett.input.adf.NumericalQuality = "Basic"
adf_exci_sett.input.adf.Symmetry = "NoSym"
adf_exci_sett.input.adf.Excitations.lowest = 10
adf_exci_sett.input.adf.Excitations.OnlySing = ""


# Import all xyz files in the folder 'molecules'

molecules = read_molecules("molecules")


print("Step 1: prescreening with DFTB")
print("==============================")

promising_molecules = {}

for name, mol in molecules.items():
    dftb_job = AMSJob(name="DFTB_" + name, molecule=mol, settings=go_sett + dftb_sett)
    dftb_job.run()

    if dftb_job.results.has_good_excitations(1, 6):
        promising_molecules[name] = dftb_job.results.get_main_molecule()

print(f"Found {len(promising_molecules)} promising molecules with DFTB")


print("Step 2: Optimization and excitations calculation with ADF")
print("=========================================================")

for name, mol in promising_molecules.items():
    adf_go_job = AMSJob(name="ADF_GO_" + name, molecule=mol, settings=go_sett + adf_sett)
    adf_go_job.run()

    optimized_mol = adf_go_job.results.get_main_molecule()

    adf_exci_job = AMSJob(name="ADF_exci_" + name, molecule=optimized_mol, settings=sp_sett + adf_exci_sett)
    adf_exci_job.run()

    if adf_exci_job.results.has_good_excitations(2, 4):
        print(f"Molecule {name} has excitation(s) satysfying our criteria!")
        print(optimized_mol)
        exci_energies, oscillator_str = adf_exci_job.results.get_excitations()
        print("Excitation energy [eV], oscillator strength:")
        for e, o in zip(exci_energies, oscillator_str):
            print(f"{e:8.4f}, {o:8.4f}")



#!/usr/bin/env python

import numpy

from ..mol.molecule import Bond, Molecule
from scm.plams.core.errors import PlamsError

__all__ = ["TrajectoryFile"]


class TrajectoryFile(object):
    """
    Abstract class that represents a generic trajectory file
    """

    def __init__(self, filename, mode="r", fileobject=None, ntap=None):
        """
        Would create a generic trajectory  file object

        * ``filename``   -- The path to the DCD file
        * ``mode``       -- The mode in which to open the DCD file ('rb' or 'wb')
        * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
        * ``ntap``       -- If the file is in write mode, the number of atoms needs to be passed here
        """
        self.position = 0
        if filename is not None:
            fileobject = open(filename, mode)
        self.file_object = fileobject
        if self.file_object is not None:
            self.mode = self.file_object.mode

        self.ntap = 0
        if ntap is not None:
            self.ntap = ntap
        self.firsttime = True
        self.coords = numpy.zeros((self.ntap, 3))  # Only for reading purposes, to avoid creating the array each time

        # PLAMS molecule related settings
        self.elements = ["H"] * self.ntap
        self.current_molecule = None
        self.store_molecule = True  # Even if True, the molecule attribute is only stored during iteration

    def __iter__(self):
        """
        Magic method that makes this an iterator

        Note: Iterates starting from current cursor position
        """
        if self.mode[0] == "r":
            if self.current_molecule is None and self.store_molecule:
                self.current_molecule = self.get_plamsmol()
            for i in range(len(self) - self.position):
                crd, cell = self.read_next(molecule=self.current_molecule)
                yield crd, cell
            self.current_molecule = None
        else:
            raise PlamsError("Cannot iterate over writable trajectoryfile")

    @property
    def molecule(self):
        """
        Returns the current molecule, which exists only when the object is used as an iterator
        """
        if not self.store_molecule:
            raise PlamsError("Try setting the store_molecule attribute to True")
        elif self.current_molecule is None:
            raise PlamsError("molecule attribute only exists during iteration")
        return self.current_molecule

    def __call__(self, molecule=None, read=True):
        """
        Magic method that makes an instance of this class into a callable
        """
        if self.mode[0] == "r":
            for i in range(len(self) - self.position):
                crd, cell = self.read_next(molecule, read)
                yield crd, cell
        else:
            raise PlamsError("Cannot iterate over writable trajectoryfile")

    def __len__(self):
        """
        Magic method that allows checking the length of an iterator
        """
        return self.get_length()

    def __enter__(self):
        """
        Magic method that allows use with *with* statement
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Magic method that allows use with *with* statement
        """
        self.close()

    def get_elements(self):
        """
        Get the elements attribute
        """
        return self.elements

    def set_elements(self, elements):
        """
        Sets the elements attribute (needed in write mode).

        *   ``elements`` -- A list containing the element symbol of each atom
        """
        self.elements = elements

    def _read_header(self):
        """
        Set up info required for reading frames

        Sets self.ntap and self.coords to proper value/size
        """
        pass

    def _move_cursor_to_append_pos(self):
        """
        Get file ready to append
        """
        filename = self.file_object.name
        mode = self.file_object.mode
        read_mode = "".join(["r", mode[1:]])

        # Read header info (requires read mode)
        # file_object = self.file_object
        self.file_object.close()
        self.file_object = open(filename, read_mode)
        self._read_header()
        self.position = self.get_length()
        self.file_object.close()

        # Reinstate in original file in append mode
        # self.file_object = file_object
        self.file_object = open(filename, mode)

    def get_plamsmol(self):
        """
        Extracts a PLAMS molecule object from the XYZ trajectory file
        """
        from scm.plams import Molecule

        oldposition = self.position
        self.rewind()
        coords, cell = self.read_next()
        plamsmol = Molecule.from_elements(self.elements)
        plamsmol.from_array(coords)

        # Return to original position
        self.rewind()
        for i in range(oldposition):
            self.read_next(read=False)

        return plamsmol

    @property
    def connection_table(self):
        """
        Symmetrize the connection table and remove bond orders
        """
        if not hasattr(self, "conect"):
            return None
        conect_sym = {}
        for i, tuples_i in self.conect.items():
            conect_sym[i] = [t[0] for t in tuples_i]
            for j, bo in tuples_i:
                if j not in conect_sym.keys():
                    conect_sym[j] = []
                if j in self.conect:
                    conect_sym[j] = [t[0] for t in self.conect[j]]
                if i not in conect_sym[j]:
                    conect_sym[j].append(i)
        return conect_sym

    def close(self):
        """
        Close the file
        """
        if hasattr(self.file_object, "close"):
            self.file_object.close()

    def read_next(self, molecule=None, read=True):
        """
        Reads the relevant info from current frame
        """
        if not read and not self.firsttime:
            return self._move_cursor_without_reading()

        # Read the coordinates and cell info from self.file_object
        self.coords = None
        self.cell = None

        # Finalize
        if self.firsttime:
            self.firsttime = False

        # Place the values into the provided molecule object
        if isinstance(molecule, Molecule):
            self._set_plamsmol(self.coords, self.cell, molecule)

        self.position += 1
        return self.coords, self.cell

    def _move_cursor_without_reading(self):
        """
        Move the cursor forward and return coords as empty array
        """
        cell = numpy.zeros((3, 3))
        if self._is_endoffile():
            return None, None
        self.position += 1
        coords = numpy.array([])
        return coords, cell

    def _is_endoffile(self):
        """
        Reads and checks If the end of file is reached.
        """
        pass

    def read_frame(self, frame, molecule=None):
        """
        Reads the relevant info from frame ``frame`` and returns it, or stores it in ``molecule``

        * ``frame``    -- The frame number to be read from the file
        * ``molecule`` -- |Molecule| object in which the new coordinates need to be stored
        """
        if frame < self.position:
            nframes = abs(frame - self.position)
            self.rewind(nframes)
        steps = frame - self.position

        for i in range(steps):
            crd, cell = self.read_next(read=False)
            if crd is None:
                break

        crd, cell = self.read_next(molecule)
        if crd is None:
            print("Not enough frames!")

        return crd, cell

    def _set_plamsmol(self, coords, cell, plamsmol):
        """
        If molecule objects have been passed as arguments, update their coordinates and lattice
        """
        plamsmol.from_array(coords)
        if cell is not None:
            if cell[0][0] ** 2 + cell[0][1] ** 2 + cell[0][2] ** 2 > 0.0:
                plamsmol.lattice = cell
                if isinstance(cell, numpy.ndarray):
                    plamsmol.lattice = cell.tolist()
        if not hasattr(self, "conect"):
            return
        if self.conect is not None:
            plamsmol.delete_all_bonds()
            bonds = []
            for i, tuples_i in self.conect.items():
                for t in tuples_i:
                    j = t
                    bo = 1.0
                    if isinstance(t, tuple):
                        j = t[0]
                        bo = t[1]
                    if sorted([i, j]) in bonds:
                        continue
                    b = Bond(plamsmol[i], plamsmol[j], bo)
                    plamsmol.add_bond(b)
                    bonds.append(sorted([i, j]))

    def write_next(self, coords=None, molecule=None, cell=[0.0, 0.0, 0.0], energy=0.0, step=None, conect=None):
        """
        Write the coordinates to the next available spot in the file
        """
        if isinstance(molecule, Molecule):
            coords, cell = self._read_plamsmol(molecule)[:2]
        cell = self._convert_cell(cell)
        # Write to self.file_object
        self.position += 1

    def _convert_cell(self, cell):
        """
        Check format of cell, and convert to vectors if needed
        """
        if cell is None:
            return cell
        # The cell can be passed as a list of three values (assumed orthorhombic)
        if len(cell) == 3 and isinstance(cell[0], float) or isinstance(cell[0], numpy.float64):
            cell = numpy.diag(cell)
        else:
            cell = numpy.array(cell)
        # For non-periodic systems there will be three cell vectors of length 0.
        if cell[0][0] ** 2 + cell[0][1] ** 2 + cell[0][2] ** 2 == 0.0:
            cell = None
        return cell

    def _read_plamsmol(self, plamsmol, read_props=True):
        """
        Read the coordinates and cell vectors from the molecule objects, if provided

        Note: For non-periodic systems the cell will be read as [0.,0.,0.]
        """
        coords = plamsmol.as_array()
        cell = [0.0, 0.0, 0.0]
        if len(plamsmol.lattice) > 0:
            cell = plamsmol.lattice
        elements = [at.symbol for at in plamsmol.atoms]
        # Get the connection table
        # plamsmol.set_atoms_id()
        conect = {}
        for bond in plamsmol.bonds:
            iat1 = plamsmol.index(bond.atom1)
            iat2 = plamsmol.index(bond.atom2)
            order = float(bond.order)
            if not iat1 in conect.keys():
                conect[iat1] = []
            conect[iat1].append((iat2, order))
            if not iat2 in conect.keys():
                conect[iat2] = []
            conect[iat2].append((iat1, order))
        # Get the properties
        props = None
        if read_props:
            from scm.plams.interfaces.adfsuite.ams import AMSJob

            props = [AMSJob._atom_suffix(at) for at in plamsmol.atoms]
            # props = [at.properties if len(at.properties)>0 else None for at in plamsmol.atoms]
            if sum([len(s) for s in props]) == 0:
                props = None
        return coords, cell, elements, conect, props

    def rewind(self, nframes=None):
        """
        Rewind the file either by ``nframes`` or to the first frame

        *   ``nframes`` -- The number of frames to rewind
        """
        if nframes is None or nframes == self.position:
            # Go back to the beginning
            self._rewind_to_first_frame()
        elif nframes > self.position:
            raise PlamsError("Trying to rewind too much!")
        else:
            # Go back nframes geometries
            self._rewind_n_frames(nframes)

    def _rewind_to_first_frame(self):
        """
        Rewind the file to the first frame
        """
        self.file_object.seek(0)
        self.firsttime = True
        self.position = 0
        self._read_header()

    def _rewind_n_frames(self, nframes):
        """
        Rewind the file by nframes frames
        """

    def get_length(self):
        """
        Get the number of frames in the file
        """
        if self.file_object.mode[0] == "a" or self.file_object.mode[0] == "w":
            return self.position
        oldposition = self.position
        while True:
            crd, cell = self.read_next(read=False)
            if crd is None:
                break
        # size = self.position-1
        size = self.position

        # Go back to the original position
        self.rewind()
        for i in range(oldposition):
            crd, cell = self.read_next(read=False)

        return size

    def read_last_frame(self, molecule=None):
        """
        Reads the last frame from the file
        """
        step = 0
        while True:
            crd, cell = self.read_next(read=False)
            if crd is None:
                break
            step += 1

        # If the current position was already the end of the file, this fails, so a failsafe is built in:
        if step == 0:
            step = self.position

        self.rewind(1)
        crd, cell = self.read_next(molecule=molecule)
        return crd, cell



import os
import pytest

import numpy as np

try:
    import dill as pickle
except ImportError:
    import pickle

from scm.plams import Atom, Molecule, MoleculeError, read_all_molecules_in_xyz_file
from scm.plams.interfaces.molecule.rdkit import from_smiles


@pytest.fixture
def BENZENE(xyz_folder):
    b = Molecule(xyz_folder / "benzene.xyz")
    b.guess_bonds()
    return b


def test_index(BENZENE):
    """Test :meth:`Molecule.index`."""
    atom = BENZENE[1]
    bond = BENZENE[1, 2]
    atom_test = Atom(coords=[0, 0, 0], symbol="H")

    assert BENZENE.index(atom) == 1
    assert BENZENE.index(bond) == (1, 2)

    try:
        BENZENE.index(None)  # None is of invalid type
    except MoleculeError:
        pass
    else:
        raise AssertionError("'BENZENE.index(None)' failed to raise a 'MoleculeError'")

    try:
        BENZENE.index(atom_test)  # atom_test is not in BENZENE
    except MoleculeError:
        pass
    else:
        raise AssertionError("'BENZENE.index(atom_test)' failed to raise a 'MoleculeError'")


def test_set_integer_bonds(BENZENE):
    """Test :meth:`Molecule.set_integer_bonds`."""
    ref1 = np.array([1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1, 1, 1, 1, 1, 1], dtype=float)
    ref2 = np.array([1, 1, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1], dtype=float)

    benzene = BENZENE.copy()
    np.testing.assert_array_equal([b.order for b in benzene.bonds], ref1)

    benzene.set_integer_bonds()
    np.testing.assert_array_equal([b.order for b in benzene.bonds], ref2)


def test_round_coords(BENZENE):
    """Test :meth:`Molecule.round_coords`."""
    benzene = BENZENE.copy()
    ref1 = np.array(
        [
            [1.0, -1.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0],
            [-1.0, 1.0, 0.0],
            [-1.0, -1.0, 0.0],
            [0.0, -1.0, 0.0],
            [2.0, -1.0, 0.0],
            [2.0, 1.0, 0.0],
            [0.0, 2.0, 0.0],
            [-2.0, 1.0, 0.0],
            [-2.0, -1.0, 0.0],
            [0.0, -2.0, 0.0],
        ]
    )
    ref2 = np.array(
        [
            [1.19, -0.69, 0.0],
            [1.19, 0.69, 0.0],
            [0.0, 1.38, 0.0],
            [-1.19, 0.69, 0.0],
            [-1.19, -0.69, 0.0],
            [-0.0, -1.38, 0.0],
            [2.13, -1.23, -0.0],
            [2.13, 1.23, -0.0],
            [0.0, 2.46, -0.0],
            [-2.13, 1.23, -0.0],
            [-2.13, -1.23, -0.0],
            [-0.0, -2.46, -0.0],
        ]
    )

    benzene2 = round(benzene)
    np.testing.assert_array_equal(benzene2, ref1)

    benzene.round_coords(decimals=2)
    np.testing.assert_allclose(benzene, ref2)


IMMUTABLE_TYPE = (int, float, tuple)
ATTR_EXCLUDE = frozenset({"mol", "bonds", "atoms", "atom1", "atom2", "_dummysymbol"})


def _compare_attrs(obj1, obj2, eval_eq=True):
    assert obj1 is not obj2

    for name, attr in vars(obj1).items():
        if name in ATTR_EXCLUDE:
            continue

        attr_ref = getattr(obj2, name)
        if eval_eq:
            assert attr == attr_ref
        if not isinstance(attr, IMMUTABLE_TYPE):
            assert attr is not attr_ref


def test_set_get_state(BENZENE, tmp_path):
    """Tests for :meth:`Molecule.__setstate__` and :meth:`Molecule.__getstate__`."""
    mol = BENZENE.copy()
    dill_new = tmp_path / "benzene_new.dill"

    with open(dill_new, "wb") as f:
        pickle.dump(mol, f)
    with open(dill_new, "rb") as f:
        mol_new = pickle.load(f)

    for mol in [mol_new, mol]:
        _compare_attrs(mol, BENZENE, eval_eq=False)

        for at, at_ref in zip(mol.atoms, BENZENE.atoms):
            _compare_attrs(at, at_ref)

        for bond, bond_ref in zip(mol.bonds, BENZENE.bonds):
            _compare_attrs(bond, bond_ref)


def test_read_multiple_molecules_from_xyz(xyz_folder):
    """Test for read_all_molecules_in_xyz_file"""

    filename = os.path.join(xyz_folder, "multiple_mols_in_xyz.xyz")

    mols = read_all_molecules_in_xyz_file(filename)

    assert len(mols) == 12

    for mol, n_atoms in zip(mols, [3, 5, 6, 3, 5, 6, 3, 5, 6]):
        assert len(mol.atoms) == n_atoms

    for mol, n_lattice_vec in zip(mols, [0, 0, 0, 0, 0, 0, 1, 2, 3, 1, 2, 3]):
        assert len(mol.lattice) == n_lattice_vec


def test_write_multiple_molecules_to_xyz(xyz_folder, tmp_path):
    """Test for append mode of Molecule.write and read_all_molecules_in_xyz_file"""

    new_xyz_file = tmp_path / "test_write_multiple_molecules.xyz"
    mols_ref = read_all_molecules_in_xyz_file(xyz_folder / "multiple_mols_in_xyz.xyz")

    assert len(mols_ref) > 1

    for mol in mols_ref:
        mol.write(new_xyz_file, mode="a")

    mols = read_all_molecules_in_xyz_file(new_xyz_file)
    assert len(mols_ref) == len(mols)

    for mol, mol_ref in zip(mols, mols_ref):
        for at, at_ref in zip(mol.atoms, mol_ref.atoms):
            assert at.symbol == at_ref.symbol
            np.testing.assert_allclose(at.coords, at_ref.coords, atol=1e-08)


def test_as_array_context():
    expected = np.array(
        [
            [-0.000816, 0.366378, -0.000000],
            [-0.812316, -0.183482, -0.000000],
            [0.813132, -0.182896, 0.000000],
        ]
    )
    mol = from_smiles("O")
    with mol.as_array as coord_array:
        np.testing.assert_allclose(coord_array, expected, rtol=1e-2)
        coord_array += 1
    np.testing.assert_allclose(mol.as_array(), expected + 1, rtol=1e-2)


def test_as_array_function():
    expected = np.array(
        [
            [-0.000816, 0.366378, -0.000000],
            [-0.812316, -0.183482, -0.000000],
            [0.813132, -0.182896, 0.000000],
        ]
    )
    mol = from_smiles("O")
    np.testing.assert_allclose(mol.as_array(), expected, rtol=1e-2)


def test_get_moments_of_inertia(BENZENE):
    benzene = BENZENE.copy()
    expected = np.array([86.81739308, 86.8173935, 173.63478658])
    np.testing.assert_allclose(benzene.get_moments_of_inertia(), expected, rtol=1e-2)


def test_get_gyration_radius(BENZENE):
    benzene = BENZENE.copy()
    expected = 1.6499992631225113
    np.testing.assert_allclose(benzene.get_gyration_radius(), expected, rtol=1e-2)


def test_separate():
    # previously this failed due to exceeding the maximum recursion depth
    # thus here we just make sure this doesn't throw an error now
    mol = Molecule(positions=[[float(i)] * 3 for i in range(1000)])
    for i in range(1, 1000):
        mol.add_bond(mol[i], mol[i + 1])
    mol.separate()



from collections import OrderedDict

from scm.plams.core.basejob import MultiJob
from scm.plams.core.functions import add_to_instance
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings
from scm.plams.recipes.md.amsmdjob import AMSNPTJob, AMSNVTJob
from scm.plams.recipes.md.scandensity import AMSMDScanDensityJob

__all__ = ["EquilibrateDensityJob", "EquilibrateDensityResults"]


class EquilibrateDensityResults(Results):
    """Results class for EquilibrateDensityJob"""

    def get_equilibrated_molecule(self, equilibration_fraction=0.667, return_index=False):
        return self.job.children["npt"].results.get_equilibrated_molecule(
            equilibration_fraction=equilibration_fraction, return_index=return_index
        )

    def rkfpath(self):
        """Returns the path to ams.rkf from the npt job"""
        return self.job.children["npt"].rkfpath()


class EquilibrateDensityJob(MultiJob):
    """A class for equilibrating the density at a certain temperature and pressure"""

    _result_type = EquilibrateDensityResults

    def _default_settings(self):
        s = Settings()
        s.input.ForceField.Type = "GAFF"
        s.input.ForceField.AnteChamberIntegration = "Yes"
        return s

    def _create_scan_density_job(self, initial_molecule):
        name = "scan_density"
        self.children[name] = AMSMDScanDensityJob(
            name=name,
            nsteps=self.nsteps[name],
            settings=self.settings,
            scan_density_upper=self.scan_density_upper,
            molecule=initial_molecule,
            **self.kwargs,
        )

        return self.children[name]

    def _create_nvt_pre_eq_job(self, scan_density_job):
        name = "nvt_pre_eq"
        job = AMSNVTJob(name=name, settings=self.settings, nsteps=self.nsteps[name], **self.kwargs)

        if scan_density_job is not None:

            @add_to_instance(job)
            def prerun(self):  # noqa F811
                self.molecule = scan_density_job.results.get_lowest_energy_molecule()

        else:
            job.molecule = self.initial_molecule

        self.children[name] = job
        return self.children[name]

    def _create_npt_job(self, nvt_pre_eq_job):
        name = "npt"
        job = AMSNPTJob.restart_from(
            nvt_pre_eq_job,
            name=name,
            use_prerun=True,
            settings=self.settings,
            nsteps=self.nsteps[name],
            **self.kwargs,
        )

        # @add_to_instance(job)
        # def prerun(self):  # noqa F811
        # self.get_velocities_from(nvt_pre_eq_job, update_molecule=True)

        self.children[name] = job
        return self.children[name]

    def __init__(
        self,
        molecule,
        settings=None,
        name="equilibrate_density",
        nsteps=None,
        scan_density=True,
        scan_density_upper=1.5,
        **kwargs,
    ):
        """
        molecule: Molecule
            3D molecule (liquid/gas with multiple molecules).

        settings: Settings
            All non-AMS-Driver settings, for example (``s.input.forcefield.type = 'GAFF'``, ``s.runscript.nproc = 1``)

        nsteps: dict
            Dictionary where the default key-values pairs are. Any keys present in the dictionary will override the default values.

            .. code-block:: python

                nsteps = {
                    'scan_density': 5000,
                    'nvt_pre_eq': 1000,
                    'npt': 100000,
                }

        kwargs: various options
            Other options for AMSMDJob (e.g. temperature, pressure, timestep)

        """
        MultiJob.__init__(self, children=OrderedDict(), name=name)

        self.scan_density_upper = scan_density_upper
        self.timestep = 1.0
        self.nsteps = {
            "scan_density": 5000,
            "nvt_pre_eq": 1000,
            "npt": 100000,
        }
        if nsteps:
            self.nsteps.update(nsteps)

        self.settings = settings.copy() if settings is not None else self._default_settings()

        self.kwargs = kwargs

        if scan_density:
            scan_density_job = self._create_scan_density_job(molecule)
        else:
            scan_density_job = None

        nvt_pre_eq_job = self._create_nvt_pre_eq_job(scan_density_job)

        self._create_npt_job(nvt_pre_eq_job)



"""
Run DFTB+ with plams
contributed by Patrick Melix
based on code from Michal Handzlik

see documentation for an example
"""

from os.path import join as opj

from scm.plams.core.basejob import SingleJob
from scm.plams.core.errors import MoleculeError
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings
from scm.plams.mol.molecule import Molecule
from scm.plams.tools.units import Units

__all__ = ["DFTBPlusJob", "DFTBPlusResults"]


class DFTBPlusResults(Results):
    """A Class for handling DFTB+ Results."""

    _outfile = "detailed.out"
    _xyzout = "geo_end.xyz"
    _genout = "geo_end.gen"

    def get_molecule(self):
        """get_molecule()
        Return the molecule from the 'geo_end.gen' file. If there is an Error, try to read the 'geo_end.xyz' file.
        """
        try:
            # The .gen file contains the cell, ASE can read it
            mol = Molecule(self[self._genout], inputformat="ase")
        except MoleculeError:
            # Fallback if no ASE found
            mol = Molecule(filename=self[self._xyzout])
        except:
            mol = Molecule()
        return mol

    def get_energy(self, string="Total energy", unit="au"):
        """get_energy(string='Total energy', unit='au')
        Return the energy given in the output with the description *string*, expressed in *unit*. Defaults to ``Total energy`` and ``au``.
        """
        try:
            energy = float(self.grep_file(self._outfile, pattern=string + ":")[0].split()[2])
            energy = Units.convert(energy, "au", unit)
        except:
            energy = float("nan")
        return energy

    def get_atomic_charges(self):
        """get_atomic_charges()
        Returns a dictonary with atom numbers and their charges, ordering is the same as in the input.
        """
        try:
            atomic_charges = {}
            string = self.awk_file(
                self._outfile, script="/Net atomic charges/{do_print=1} NF==0 {do_print=0 }do_print==1 {print}"
            )
            for line in string:
                if string[0] == line or string[1] == line:
                    continue
                l = line.split()
                atomic_charges[l[0]] = float(l[1])
        except:
            atomic_charges = {}
        return atomic_charges


class DFTBPlusJob(SingleJob):
    """A class representing a single computational job with DFTB+.
    Only supports molecular coordinates, no support for lattice yet."""

    _result_type = DFTBPlusResults
    _filenames = {"inp": "dftb_in.hsd", "run": "$JN.run", "out": "$JN.out", "err": "$JN.err", "gen": "$JN.gen"}

    def get_input(self):
        """Transform all contents of ``setting.input`` branch into string with blocks, keys and values.

        Automatic handling of ``molecule`` can be disabled with ``settings.ignore_molecule = True``.
        """

        def parse(key, value, indent=""):
            if value is True:
                value = ""

            ret = indent + key
            if key != "" and value != "":
                ret += " ="

            if isinstance(value, Settings):
                if "_h" in value:
                    ret += " " + value["_h"]
                ret += " {\n"

                i = 1
                while ("_" + str(i)) in value:
                    ret += parse("", value["_" + str(i)], indent + "  ")
                    i += 1

                for el in value:
                    if not el.startswith("_"):
                        ret += parse(el, value[el], indent + "  ")
                ret += indent + "}"
            else:
                ret += " " + str(value)
            ret += "\n"
            return ret

        inp = ""
        use_molecule = ("ignore_molecule" not in self.settings) or (self.settings.ignore_molecule == False)
        if use_molecule:
            self._parsemol()

        for item in self.settings.input:
            inp += parse(item, self.settings.input[item])

        if use_molecule:
            self._removemol()
        return inp

    def _parsemol(self):
        # use ASE to write molecule if available
        if "ase" in Molecule._writeformat:
            filename = opj(self.path, self._filename("gen"))
            self.molecule.write(filename, outputformat="ase", format="gen")
            self.settings.input.geometry._h = "GenFormat"
            self.settings.input.geometry._1 = "<<< " + self._filename("gen")

        else:
            # Old way of handling gen-format ourselves, delete if ASE becomes obligatory
            atom_types = {}
            n = 1
            atoms_line = ""
            for atom in self.molecule:
                if atom.symbol not in atom_types:
                    atoms_line += atom.symbol + " "
                    atom_types[atom.symbol] = n
                    n += 1

            # check PBC
            lattice = []
            geomType = "C"
            for vec in self.molecule.lattice:
                if not all(isinstance(x, (int, float)) for x in vec):
                    raise ValueError("Non-Number in Lattice Vectors, not compatible with DFTBPlus")

                lattice.append(vec)
                geomType = "S"

            self.settings.input.geometry._h = "GenFormat"
            self.settings.input.geometry._1 = "%i %s" % (len(self.molecule), geomType)
            self.settings.input.geometry._2 = atoms_line
            self.settings.input.geometry._3 = ""

            for i, atom in enumerate(self.molecule):
                self.settings.input.geometry["_" + str(i + 4)] = ("%5i" % (i + 1)) + atom.str(
                    symbol=str(atom_types[atom.symbol])
                )

            if len(vec) > 0:
                j = i + 1
                # origin
                self.settings.input.geometry["_" + str(j + 4)] = "0.0 0.0 0.0"
                j += 1
                for i, vec in enumerate(lattice):
                    self.settings.input.geometry["_" + str(i + j + 4)] = "%f %f %f" % (vec)

    def _removemol(self):
        if "geometry" in self.settings.input:
            del self.settings.input.geometry

    def get_runscript(self):
        """dftb+ has to be in your $PATH!"""
        ret = "dftb+ "
        if self.settings.runscript.stdout_redirect:
            ret += " >" + self._filename("out")
        ret += "\n\n"
        return ret

    def check(self):
        """Returns true if 'ERROR!' is not found in the output."""
        s = self.results.grep_output("ERROR!")
        return len(s) == 0
